/* SWR frame/hitbox display tool
 * Developed on Linux - I don't really care about Windows support.
 *
 * written by mauve of irc.mizuumi.net #iamp
 * solid box rendering code by tropsy
 *
 * this code is unsupported.
 *
 * it is also extremely sloppy.
 *
 * compile with MinGW32 and libSDL
 * g++ -o framedisplayswr framedisplayswr.cpp -lmingw32 -lSDLmain -lSDL -lopengl32
 */

#include <SDL.h>
#include <SDL_keysym.h>
#include <GL/gl.h>
#include <GL/glext.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <map>
#include <string>

#include "framedisplayfont.h"

#ifdef WIN32
#include <windows.h>
#endif

// th105a.dat/th105c.dat's headers are encrypted with a mersenne twister

class MersenneTwister {
  private:
    unsigned long m_mt[624];
    unsigned int m_v;

    void generate_next_set() {
	static const unsigned long table[2]
	= { 0, 2567483615UL };
	unsigned long v;

	for (int i = 0; i < 623; ++i) {
	    v = (m_mt[i] & 0x80000000) | (m_mt[i + 1] & 0x7fffffff);

	    m_mt[i] = m_mt[(i + 397) % 624] ^ (v >> 1) ^ table[v & 1];
	}

	v = (m_mt[623] & 0x80000000) | (m_mt[0] & 0x7fffffff);
	m_mt[623] = m_mt[396] ^ (v >> 1);
	if (v & 1) {
	    m_mt[623] ^= table[1];
	}

	m_v = 0;
    }
  public:
    void seed(unsigned long seed_v) {
	m_mt[0] = seed_v;
	for (int i = 1; i < 624; ++i) {
	    m_mt[i] =
		1812433253UL * (m_mt[i - 1] ^ (m_mt[i - 1] >> 30)) + i;
	}
	generate_next_set();
    }

    unsigned long rand_long() {
	if (m_v >= 624) {
	    generate_next_set();
	}

	unsigned long v = m_mt[m_v++];

	v ^= (v >> 11);
	v ^= (v << 7) & 0x9d2c5680;
	v ^= (v << 15) & 0xefc60000;
	v ^= (v >> 18);

	return v;
    }

    MersenneTwister() {
	m_v = 0;
    }

    MersenneTwister(unsigned long seed_v) {
	m_v = 0;

	seed(seed_v);
    }
};

struct datfile_t;
struct datentry_t;

struct image_slot_t {
    bool valid;

    char filename[128];

    int width;
    int height;
    int pitch;
    float x_mult;
    float y_mult;

    datfile_t *datfile;
    datentry_t *datentry;

    bool is_8bpp;

    unsigned char *pixels;
    GLuint tex_id;

    int tex_cycle;
};

struct rect_t {
    int x1, y1, x2, y2;
};

struct frame_t {
    struct {
	short render_frame;
	short unk2;
	short unk3;
	short unk4;
	short tex_width;	// x2?
	short tex_height;	// y2?
	short x_offset;
	short y_offset;
	short duration;
	unsigned char identifier;
    } __attribute__ ((packed)) header;

    struct {
	short blend_mode;
	short unk2;
	short unk3;
	short scale;
	short unk5;
	short unk6;
	short angle;
    } __attribute__ ((packed)) type_2;
    
    struct {
	short blend_mode;
	short unk2;
	short unk3;
	short unk123_1;
	short scale;
	short unk5;
	short unk6;
	short angle;
    } __attribute__ ((packed)) th123_type_2;
    
    struct {
	short damage;
	short proration;
	short unk3;
	short unk4;
	short untech;
	short newunk5;
	short limit;
	short unk5;
	short unk6;
	short unk7;
	short unk8;
	short unk9;
	short unk10;
	short unk11;
	short unk12;
	short velocity_x;
	short velocity_y;
	short unk15;
	short unk16;
	short type;
	unsigned char unk18;
	int fflags;
	int aflags;
    } __attribute__ ((packed)) data;

    int n_rects;
    rect_t *rects;

    int n_selfboxes;
    rect_t *selfboxes;

    int n_hitboxes;
    rect_t *hitboxes;

    struct {
        short unk[0xf];
    } __attribute__ ((packed)) th123_block;
};

struct sequence_t {
    int identifier;
    struct {
	short unk1;
	short unk2;

	unsigned char unk3;

	int nframes;
    } __attribute__ ((packed)) header;
    struct {
	int unk1;
	int unk2;
    } alt_header;

    frame_t *frames;
};

struct character_t {
    int nsequences;
    sequence_t *sequences;

    struct {
        unsigned int *data;
        bool active;
    } palettes[10];

    char *name;
    
    bool loaded;

    int nimages;
    image_slot_t *images;
};

char **character_names = 0;
character_t *characters = 0;
int ncharacters = 0;

struct datentry_t {
    char filename[256];
    unsigned char xorkey;
    unsigned int size;
    unsigned int position;
};

struct datfile_t {
    char *filename;

    short nfiles;

    struct datentry_t *files;
     std::map < std::string, int >file_map;
};

datfile_t *dat_open(const char *filename)
{
    FILE *file = fopen(filename, "rb");
    if (!file) {
	return 0;
    }

    datfile_t *datfile = new datfile_t;
    unsigned long size;

    fread(&datfile->nfiles, 2, 1, file);
    fread(&size, 4, 1, file);
    char *index_buf = new char[size];
    char *orig_index_buf = index_buf;

    MersenneTwister prng;

    prng.seed(size + 6);

    datfile->filename = strdup(filename);

    fread(index_buf, size, 1, file);

    fclose(file);

    for (unsigned int i = 0; i < size; ++i) {
	unsigned char a = prng.rand_long();
	index_buf[i] ^= a;
    }

    {
	unsigned char xorval = 0xc5;
	unsigned char xoradd = 0x83;
	unsigned char xoraddadd = 0x53;

	for (unsigned int i = 0; i < size; ++i) {
	    index_buf[i] ^= xorval;
	    xorval += xoradd;
	    xoradd += xoraddadd;
	}

    }

    datfile->files = new datentry_t[datfile->nfiles];

    for (int i = 0; i < datfile->nfiles; ++i) {
	datfile->files[i].position = *(int *) (index_buf);
	datfile->files[i].size = *(int *) (index_buf + 4);

	unsigned char string_len = *(unsigned char *) (index_buf + 8);
	memcpy(datfile->files[i].filename, index_buf + 9, string_len);
	datfile->files[i].filename[string_len] = '\0';

	datfile->files[i].xorkey =
	    ((datfile->files[i].position >> 1) | 0x23) & 0xff;

	datfile->file_map[datfile->files[i].filename] = i + 1;

	index_buf += string_len + 9;
    }

    delete[]orig_index_buf;

    return datfile;
}

void memxor(char *data, unsigned char value, unsigned int length)
{
    for (unsigned int i = 0; i < length; ++i) {
	data[i] ^= value;
    }
}

char *dat_read_direct(datfile_t * datfile, int i, int *dsize)
{
    if (!datfile) {
	return 0;
    }

    FILE *file = fopen(datfile->filename, "rb");
    fseek(file, datfile->files[i].position, SEEK_SET);

    char *data = new char[datfile->files[i].size];
    fread(data, datfile->files[i].size, 1, file);

    fclose(file);

    memxor(data, datfile->files[i].xorkey, datfile->files[i].size);

    *dsize = datfile->files[i].size;

    return data;
}

int dat_find(datfile_t * datfile, const char *filename)
{
    if (!datfile) {
	return -1;
    }

    int n = datfile->file_map[filename];

    return n - 1;
}

char *dat_read(datfile_t * datfile, const char *filename, int *dsize)
{
    int i = dat_find(datfile, filename);
    if (i < 0) {
	return 0;
    }

    return dat_read_direct(datfile, i, dsize);
}

FILE *dat_openfile(datfile_t * datfile, const char *filename, int *dsize)
{
    int i;
    for (i = 0; i < datfile->nfiles; ++i) {
	if (!strcmp(datfile->files[i].filename, filename)) {
	    break;
	}
    }

    if (i == datfile->nfiles) {
	return 0;
    }

    FILE *file = fopen(datfile->filename, "rb");
    fseek(file, datfile->files[i].position, SEEK_SET);

    *dsize = datfile->files[i].size;

    return file;
}

void read_pat(datfile_t * dat_105a, datfile_t *dat_123a, datfile_t *dat_105c, datfile_t *dat_123c,
              bool th123,
	      character_t * character, char *data, char *end)
{
    // new to swr, first up we've got a version number, then two numbers
    // indicating the number of images.
    unsigned char pat_version = *data++;
    if (pat_version != 3) {
	// umm
    }
    
    // read the images in.
    unsigned short image_count = *((unsigned short *) data);
    data += 2;

    character->nimages = image_count;
    character->images = new image_slot_t[image_count];
    memset(character->images, 0, sizeof(image_slot_t) * image_count);

    int nimages = 0;
    image_slot_t *image = character->images;

    while (nimages < image_count) {
	char filename[256];
	memcpy(filename, data, 128);
	data += 128;
	filename[128] = '\0';

	char *bmp = strstr(filename, ".bmp");
	if (bmp) {
	    *bmp = '\0';
	}

	sprintf(image->filename, "data/character/%s/%s.cv2",
		character->name, filename);

	image->valid = 0;

	int n = dat_find(dat_123c, image->filename);
        if (n >= 0) {
            image->datfile = dat_123c;
            image->datentry = &dat_123c->files[n];
            image->valid = 1;
        } else if ((n = dat_find(dat_123a, image->filename)) >= 0) {
            image->datfile = dat_123a;
            image->datentry = &dat_123a->files[n];
            image->valid = 1;
        } else if ((n = dat_find(dat_105c, image->filename)) >= 0) {
            image->datfile = dat_105c;
            image->datentry = &dat_105c->files[n];
            image->valid = 1;
        } else if ((n = dat_find(dat_105a, image->filename)) >= 0) {
            image->datfile = dat_105a;
            image->datentry = &dat_105a->files[n];
            image->valid = 1;
        }

	image++;
	nimages++;
    }
    
    // now, read the frame data in.
    sequence_t *seq;

    int seq_count;

    seq_count = *((int *) data);
    data += 4;

    character->nsequences = seq_count;
    character->sequences = new sequence_t[seq_count];

    memset(character->sequences, 0, seq_count * sizeof(sequence_t));

    seq = character->sequences;

    int nseq, nframe;

    nseq = 0;
    
    while (seq_count > 0 && data < end) {
	seq_count--;
	memcpy(&seq->identifier, data, sizeof(seq->identifier));
	data += sizeof(seq->identifier);

	if (seq->identifier == -1) {
	    //memcpy(&seq->alt_header, data, sizeof(seq->alt_header));
	    data += sizeof(seq->alt_header);

	    //seq->header.nframes = 0;
	    //seq->frames = 0;

	    //seq++;
	    character->nsequences--;

	    continue;
	}

	nseq++;


	memcpy(&seq->header, data, sizeof(seq->header));
	data += sizeof(seq->header);

	int frame_count = seq->header.nframes;

	seq->frames = new frame_t[frame_count];

	memset(seq->frames, 0, sizeof(frame_t) * frame_count);

	frame_t *frame = seq->frames;

	nframe = 0;
	
	//printf("%d: %5.5x\n",frame_count, data - start);

	while (frame_count > 0) {
	    nframe++;
	    frame_count--;

	    /*
	    for (int i = 0; i < sizeof(frame->header) ; ++i) {
	        printf("%2.2x", (unsigned char)data[i]);
	    }
	     */
	    memcpy(&frame->header, data, sizeof(frame->header));
	    data += sizeof(frame->header);
	    
	    //printf(" %d",frame->header.identifier);

	    if (frame->header.identifier == 2) {
		if (!th123) {
		  memcpy(&frame->type_2, data, sizeof(frame->type_2));
		  data += sizeof(frame->type_2);
		} else {
		  memcpy(&frame->th123_type_2, data, sizeof(frame->th123_type_2));
		  data += sizeof(frame->th123_type_2);
		  
		  frame->type_2.blend_mode = frame->th123_type_2.blend_mode;
		  frame->type_2.unk2 = frame->th123_type_2.unk2;
		  frame->type_2.unk3 = frame->th123_type_2.unk3;
		  frame->type_2.scale = frame->th123_type_2.scale;
		  frame->type_2.unk5 = frame->th123_type_2.unk5;
		  frame->type_2.unk6 = frame->th123_type_2.unk6;
		  frame->type_2.angle = frame->th123_type_2.angle;
                }
	    }

	    memcpy(&frame->data, data, sizeof(frame->data));
	    data += sizeof(frame->data);

	    frame->n_rects = *((char *) data);
	    //printf(": %d, ",frame->n_rects);
	    data += 1;
	    if (frame->n_rects > 0) {
		frame->rects = new rect_t[frame->n_rects];
		memcpy(frame->rects, data,
		       sizeof(rect_t) * frame->n_rects);
		data += sizeof(rect_t) * frame->n_rects;
	    }

	    frame->n_selfboxes = *((char *) data);
	    //printf(": %d, ",frame->n_selfboxes);
	    data += 1;
	    if (frame->n_selfboxes > 0) {
		frame->selfboxes = new rect_t[frame->n_selfboxes];
		memcpy(frame->selfboxes, data,
		       sizeof(rect_t) * frame->n_selfboxes);
		data += sizeof(rect_t) * frame->n_selfboxes;
	    }

	    frame->n_hitboxes = *((char *) data);
	    //printf(": %d : ",frame->n_hitboxes);
	    data += 1;
	    if (frame->n_hitboxes > 0) {
		frame->hitboxes = new rect_t[frame->n_hitboxes];

		for (int j = 0; j < frame->n_hitboxes; ++j) {
		    memcpy(&frame->hitboxes[j], data, sizeof(rect_t));
		    data += sizeof(rect_t);
		    data++;
		}
		//memcpy(frame->hitboxes, data,
		//       sizeof(rect_t) * frame->n_hitboxes);
		//data += sizeof(rect_t) * frame->n_hitboxes;
	    }

	    /*
	    for (int i = 0; i < 0x1e; ++i) {
	        printf("%2.2x",data[i]);
	    }
	    printf("\n");
	     */
	    if (th123) {
	        memcpy(&frame->th123_block, data, 0x1e);
                data += 0x1e;
	    }

	    frame++;
	}

	seq++;
    }

    character->nsequences = nseq;
}

void display_load();

void read_pal(datfile_t * dat_105a, datfile_t * dat_123a, datfile_t * dat_105c, datfile_t * dat_123c, int ch_no)
{
    for (int j = 0; j < 10; ++j) {
	char filename[256];
	sprintf(filename, "data/character/%s/palette%3.3d.pal",
		character_names[ch_no], j);

	int size;
	char *data = dat_read(dat_123c, filename, &size);
	if (!data) {
	    data = dat_read(dat_123a, filename, &size);
	}
	if (!data) {
	    data = dat_read(dat_105c, filename, &size);
	}
	if (!data) {
	    data = dat_read(dat_105a, filename, &size);
	}
	unsigned short *pal_src = (unsigned short *) (data + 1);
	unsigned int *palette = new unsigned int[256];

	characters[ch_no].palettes[j].data = palette;
	characters[ch_no].palettes[j].active = 0;

	if (!data) {
	    palette[0] = 0;
	    for (int j = 1; j < 256; ++j) {
	        palette[j] = 0xffffffff;
	    }
	    
	    continue;
	}
	
	characters[ch_no].palettes[j].active = 1;
	
	for (int j = 0; j < 256; ++j) {
	    unsigned long v = *pal_src++;
	    unsigned long o;

	    o = (v & 0x8000) ? 0xff000000 : 00;
	    o |= ((v & 0x7c00) << 1) >> 8;
	    o |= ((v & 0x03e0) << 6);
	    o |= ((v & 0x001f) << 3) << 16;

	    palette[j] = o;
	}
	
	delete[] data;
	
	if (!strcmp(character_names[ch_no], "utsuho")) {
	    if (palette[31] == 0xfff800f8) {
	        palette[31] = 0xff201010;
            }
	}
    }
}

void read_pats(datfile_t * dat_105a, datfile_t * dat_123a, datfile_t * dat_105c, datfile_t *dat_123c)
{
    int j = 0;

    datfile_t *d = dat_123a;
    if (!d) {
        d = dat_105a;
    }
    ncharacters = 0;
    for (int i = 0; i < d->nfiles; ++i) {
	int l = strlen(d->files[i].filename);

	if (!memcmp(d->files[i].filename, "data/character/", 15)
	    && !strcmp(d->files[i].filename + l - 4, ".pat")) {
	    ncharacters++;
	}
    }
    if (ncharacters == 0) {
	return;
    }
    
    int oldnch = ncharacters;

    character_names = new char *[ncharacters];
    characters = new character_t[ncharacters];
    memset(characters, 0, sizeof(character_t) * ncharacters);

    ncharacters = 0;

    for (int k = 0; k < 4; ++k) {
	datfile_t *d = 0;
	bool th123;
	switch(k) {
	case 0: d = dat_123c; th123 = 1; break;
	case 1: d = dat_123a; th123 = 1; break;
	case 2: d = dat_105c; th123 = 0; break;
	case 3: d = dat_105a; th123 = 0; break;
	}

	if (!d) {
	    continue;
	}

	for (int i = 0; i < d->nfiles; ++i) {
	    datentry_t *entry = &d->files[i];
	    int l = strlen(entry->filename);
	    if (!strncmp(entry->filename, "data/character/", 15)
		&& !strcmp(entry->filename + l - 4, ".pat")) {
		char *end = strchr(entry->filename + 15, '/');
		if (!end) {
		    ncharacters = 0;
		    return;
		}

		int len = end - (entry->filename + 15);
		char *name = new char[len + 1];
		memcpy(name, entry->filename + 15, len);
		name[len] = '\0';

		if (k == 1) {
		    int l;
		    for (l = 0; l < j; ++l) {
			if (!strcmp(character_names[l], name)) {
			    free(name);
			    break;
			}
		    }
		    if (l != j) {
			continue;
		    }
		}
		
                character_names[j] = name;
                characters[j].name = name;

		int size;
		char *pat_data = dat_read_direct(d, i, &size);

		if (pat_data) {
		    if (!characters[j].loaded) {
		        read_pat(dat_105a, dat_123a, dat_105c, dat_123c, th123, &characters[j], pat_data,
			     pat_data + size);
                        read_pal(dat_105a, dat_123a, dat_105c, dat_123c, j);
                        ncharacters++;
		        characters[j].loaded = 1;
		        display_load();
		        
		        j++;
                    }
		    delete[]pat_data;
		    
		    if (j == oldnch) {
		        break;
		    }
		}
	    }
	}
	
	if (j == oldnch) {
	    break;
	}
    }
}

int to_pow_2(int v)
{
    int a = 1;

    while (a < v) {
	a <<= 1;
    }

    return a;
}

static int current_tex_cycle = 1;
int alt_colors = 0;

void flush_images()
{
    for (int i = 0; i < ncharacters; ++i) {
	character_t *ch = &characters[i];
	for (int j = 0; j < ch->nimages; ++j) {
	    image_slot_t *image = &ch->images[j];
	    if (image->tex_cycle != current_tex_cycle || !image->pixels) {
		continue;
	    }

	    if (!image->is_8bpp) {
		continue;
	    }

	    glDeleteTextures(1, &image->tex_id);

	    image->pixels = 0;
	}
    }
}

void read_image(character_t * character, image_slot_t * image)
{
    if (!image->datfile) {
        return;
    }
    FILE *file = fopen(image->datfile->filename, "rb");
    fseek(file, image->datentry->position, SEEK_SET);

    struct {
	unsigned char type;
	unsigned int width;
	unsigned int height;
	unsigned int pitch;
	unsigned char unknown;
	short unknown2;
	unsigned char unknown3;
    } __attribute__ ((packed)) header;

    fread(&header, 17, 1, file);

    unsigned char xorkey = image->datentry->xorkey;

    memxor((char *) &header, xorkey, 17);

    int width = header.width;
    int pitch = header.pitch;
    int height = header.height;
    image->width = width;
    image->height = height;

    if (image->width < 0 || image->height < 0 || image->width > 2048
	|| image->height > 2048) {
	return;
    }

    int rwidth = to_pow_2(image->width);
    int rheight = to_pow_2(image->height);
    int count = width * height;
    int rcount = rwidth * rheight;
    int x;

    unsigned int *pixels = new unsigned int[rcount];
    memset(pixels, 0, rcount * 4);

    unsigned int *p = pixels;
    unsigned int pixel;
    image->pixels = (unsigned char *) pixels;
    image->is_8bpp = 0;

    if (header.type != 0x8) {
	x = 0;
	while (count > 0) {
	    fread(&pixel, 4, 1, file);
	    memxor((char *) &pixel, xorkey, 4);

	    unsigned int npixel = pixel;

	    npixel = pixel & 0xff00ff00;
	    npixel |= (pixel & 0xff0000) >> 16;
	    npixel |= (pixel & 0x0000ff) << 16;

	    *p++ = npixel;
	    count--;
	    x++;
	    if (x == width) {
		p += rwidth - width;
		for (int i = 0; i < (pitch - width); ++i) {
		    fread(&pixel, 4, 1, file);
		}
		x = 0;
	    }
	    if (count <= 0) {
		break;
	    }
	}
    } else {
	image->is_8bpp = 1;
	x = 0;
	pixel = 0;
	unsigned int *palette = character->palettes[alt_colors].data;
	if (!character->palettes[alt_colors].active) {
	    for (int i = 0; i < 10; ++i) {
	        if (character->palettes[i].active) {
	            palette = character->palettes[i].data;
	            break;
	        }
	    }
	}
	while (count > 0) {
	    fread(&pixel, 1, 1, file);
	    pixel ^= xorkey;

	    unsigned int npixel = palette[pixel];

	    *p++ = npixel;
	    count--;
	    x++;
	    if (x == width) {
		p += rwidth - width;
		for (int i = 0; i < (pitch - width); ++i) {
		    fread(&pixel, 1, 1, file);
		}
		x = 0;
	    }
	    if (count <= 0) {
		break;
	    }
	}
    }

    fclose(file);

    GLuint tex_id;

    glGenTextures(1, &tex_id);
    image->tex_id = tex_id;

    glBindTexture(GL_TEXTURE_2D, tex_id);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, rwidth, rheight, 0, GL_RGBA,
		 GL_UNSIGNED_BYTE, pixels);

    image->x_mult = (float) width / (float) rwidth;
    image->y_mult = (float) height / (float) rheight;

    image->tex_cycle = current_tex_cycle;

    delete[]pixels;
}

// tropsy's code begins here

struct node {
    int x1, y1, x2, y2;
    node *next;
};
node *start_ptr = NULL;

node *create_box(frame_t * frame, int type)
{				// type 0 = selfboxes, 1 = hitboxes
    node *first = NULL;
    node *cur = NULL;
    int error;
    if (type == 0) {
	for (int i = 0; i < frame->n_selfboxes; ++i) {
	    node *tmp = new node;
	    tmp->next = NULL;
	    if (cur != NULL) {
		cur->next = tmp;
	    }
	    if (first == NULL) {
		first = tmp;
	    }
	    cur = tmp;
	    rect_t *rect = &frame->selfboxes[i];;
	    cur->x1 = rect->x1;
	    cur->x2 = rect->x2;
	    cur->y1 = rect->y1;
	    cur->y2 = rect->y2;
	}
    } else if (type == 1) {
	for (int i = 0; i < frame->n_hitboxes; ++i) {
	    node *tmp = new node;
	    tmp->next = NULL;
	    if (cur != NULL) {
		cur->next = tmp;
	    }
	    if (first == NULL) {
		first = tmp;
	    }
	    cur = tmp;
	    rect_t *rect = &frame->hitboxes[i];
	    cur->x1 = rect->x1;
	    cur->x2 = rect->x2;
	    cur->y1 = rect->y1;
	    cur->y2 = rect->y2;
	}
    } else {
	for (int i = 0; i < frame->n_rects; ++i) {
	    node *tmp = new node;
	    tmp->next = NULL;
	    if (cur != NULL) {
		cur->next = tmp;
	    }
	    if (first == NULL) {
		first = tmp;
	    }
	    cur = tmp;
	    rect_t *rect = &frame->rects[i];
	    cur->x1 = rect->x1;
	    cur->x2 = rect->x2;
	    cur->y1 = rect->y1;
	    cur->y2 = rect->y2;
	}
    }
    do {
	error = 0;
	cur = first;
	while (cur->next != NULL) {
	    node *tmp = NULL;
	    do {
		if (tmp == NULL) {
		    tmp = cur->next;
		} else {
		    tmp = tmp->next;
		}
		if ((cur->x1 < tmp->x1) && (cur->x2 > tmp->x1)) {
		    node *new1 = new node;
		    node *new2 = new node;
		    new1->x1 = cur->x1;
		    new1->x2 = tmp->x1;
		    new2->x1 = tmp->x1;
		    new2->x2 = cur->x2;
		    new1->y1 = cur->y1;
		    new1->y2 = cur->y2;
		    new2->y1 = cur->y1;
		    new2->y2 = cur->y2;
		    new1->next = new2;
		    new2->next = cur->next;
		    node *tmp2 = first;
		    if (cur == first) {
			first = new1;
			delete tmp2;
		    } else {
			while (tmp2->next != cur) {
			    tmp2 = tmp2->next;
			}
			tmp2->next = new1;
			delete cur;
			cur = first;
		    }
		    error = 1;
		    break;
		}
		if ((cur->x1 < tmp->x2) && (cur->x2 > tmp->x2)) {
		    node *new1 = new node;
		    node *new2 = new node;
		    new1->x1 = cur->x1;
		    new1->x2 = tmp->x2;
		    new2->x1 = tmp->x2;
		    new2->x2 = cur->x2;
		    new1->y1 = cur->y1;
		    new1->y2 = cur->y2;
		    new2->y1 = cur->y1;
		    new2->y2 = cur->y2;
		    new1->next = new2;
		    new2->next = cur->next;
		    node *tmp2 = first;
		    if (cur == first) {
			first = new1;
			delete tmp2;
		    } else {
			while (tmp2->next != cur) {
			    tmp2 = tmp2->next;
			}
			tmp2->next = new1;
			delete cur;
			cur = first;
		    }
		    error = 1;
		    break;
		}
		if ((cur->y1 < tmp->y1) && (cur->y2 > tmp->y1)) {
		    node *new1 = new node;
		    node *new2 = new node;
		    new1->y1 = cur->y1;
		    new1->y2 = tmp->y1;
		    new2->y1 = tmp->y1;
		    new2->y2 = cur->y2;
		    new1->x1 = cur->x1;
		    new1->x2 = cur->x2;
		    new2->x1 = cur->x1;
		    new2->x2 = cur->x2;
		    new1->next = new2;
		    new2->next = cur->next;
		    node *tmp2 = first;
		    if (cur == first) {
			first = new1;
			delete tmp2;
		    } else {
			while (tmp2->next != cur) {
			    tmp2 = tmp2->next;
			}
			tmp2->next = new1;
			delete cur;
			cur = first;
		    }
		    error = 1;
		    break;
		}
		if ((cur->y1 < tmp->y2) && (cur->y2 > tmp->y2)) {
		    node *new1 = new node;
		    node *new2 = new node;
		    new1->y1 = cur->y1;
		    new1->y2 = tmp->y2;
		    new2->y1 = tmp->y2;
		    new2->y2 = cur->y2;
		    new1->x1 = cur->x1;
		    new1->x2 = cur->x2;
		    new2->x1 = cur->x1;
		    new2->x2 = cur->x2;
		    new1->next = new2;
		    new2->next = cur->next;
		    node *tmp2 = first;
		    if (cur == first) {
			first = new1;
			delete tmp2;
		    } else {
			while (tmp2->next != cur) {
			    tmp2 = tmp2->next;
			}
			tmp2->next = new1;
			delete cur;
			cur = first;
		    }
		    error = 1;
		    break;
		}
		if ((cur->x1 >= tmp->x1) && (cur->x2 <= tmp->x2)
		    && (cur->y1 >= tmp->y1) && (cur->y2 <= tmp->y2)) {
		    node *tmp2 = first;
		    if (cur == first) {
			first = first->next;
			delete tmp2;
		    } else {
			while (tmp2->next != cur) {
			    tmp2 = tmp2->next;
			}
			tmp2->next = cur->next;
			tmp2 = cur;
			cur = first;
			delete tmp2;
		    }
		    error = 1;
		    break;
		}
	    }
	    while (tmp->next != NULL);
	    if (error == 1) {
		break;
	    }
	    cur = cur->next;
	}
    }
    while (error != 0);
    return first;
}

// tropsy's code ends here

int draw_gradient = 1;
int draw_properties = 1;
int draw_sprite = 1;
int draw_collisionbox = 1;
int draw_hitbox = 1;
int draw_damagebox = 1;
int draw_additive = 1;
int pretty_display = 1;
int draw_boxes = 1;
int realframes = 0;
int draw_axises = 1;

int position_x = 180;
int position_y = 400;

int window_width = 640;
int window_height = 480;
float window_scale = 1.0;

static GLuint font_tex_id = 0;

void init_font()
{
    unsigned char data[256 * 64];
    for (int i = 0; i < 256 * 64; ++i) {
	data[i] = 255 - header_data_cmap[header_data[i]][0];
    }

    glGenTextures(1, &font_tex_id);

    glBindTexture(GL_TEXTURE_2D, font_tex_id);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA8, 256, 64, 0, GL_ALPHA,
		 GL_UNSIGNED_BYTE, data);
}

void draw_font(int dx, int dy, const char *string)
{
    int x = dx;
    int y = dy;

    const float c_w = 10.0 / 256.0;
    const float c_h = 15.0 / 64.0;

    glBindTexture(GL_TEXTURE_2D, font_tex_id);
    glEnable(GL_TEXTURE_2D);
    glBegin(GL_QUADS);

    while (*string) {
	char ch = *string++;
	int c;

	if (ch >= 'A' && ch <= 'Z') {
	    c = ch - 'A';
	} else if (ch >= 'a' && ch <= 'z') {
	    c = 26 + ch - 'a';
	} else if (ch >= '0' && ch <= '9') {
	    c = 52 + ch - '0';
	} else if (ch == '-') {
	    c = 52 + 10;
	} else if (ch == '\n') {
	    x = dx;
	    y += 13;
	    continue;
	} else if (ch == ' ') {
	    x += 9;
	    continue;
	} else {
	    continue;
	}

	float cx = (float) (c % 25) * 10.0 / 256.0;
	float cy = (float) (c / 25) * 15.0 / 64.0;

	// i messed up, too lazy to fix
	if (c == 50 || c == 51) {
	    cy -= 1.0 / 64.0;
	}

	glTexCoord2f(cx, cy);
	glVertex2f(x, y);
	glTexCoord2f(cx + c_w, cy);
	glVertex2f(x + 10, y);
	glTexCoord2f(cx + c_w, cy + c_h);
	glVertex2f(x + 10, y + 15);
	glTexCoord2f(cx, cy + c_h);
	glVertex2f(x, y + 15);

	x += 9;
    }

    glEnd();

    glDisable(GL_TEXTURE_2D);
}

void display_bg()
{
    glBegin(GL_QUADS);
    glColor4f(0.1, 0.05, 0.0, 1.0);
    glVertex2f(0.0, 0.0);
    glVertex2f(window_width, 0.0);
    if (draw_gradient) {
	glColor4f(0.2, 0.1, 0.0, 1.0);
    }
    glVertex2f(window_width, window_height);
    glVertex2f(0.0, window_height);
    glEnd();
}

PFNGLBLENDEQUATIONEXTPROC pglBlendEquation = 0;

void display_sprite(character_t * character, frame_t * frame)
{
    if (frame->header.render_frame < 0
	|| frame->header.render_frame >= character->nimages) {
	return;
    }
    image_slot_t *image = &character->images[frame->header.render_frame];

    if (!image->pixels || image->tex_cycle != current_tex_cycle) {
	read_image(character, image);
    }

    if (image->pixels) {
	int x1, y1, x2, y2;

	glPushMatrix();

	if (frame->header.identifier == 2) {
	    if (frame->type_2.blend_mode == 1 && draw_additive) {
		if (pglBlendEquation) {
		    pglBlendEquation(GL_FUNC_ADD);
		}
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	    } else if (frame->type_2.blend_mode == 2 && draw_additive) {
		if (pglBlendEquation) {
		    pglBlendEquation(GL_FUNC_REVERSE_SUBTRACT);
		}
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	    }

	    float scale = ((float) frame->type_2.scale) / 100.0;

	    x1 = -(int) ((float) frame->header.x_offset * scale);
	    y1 = -(int) ((float) frame->header.y_offset * scale);
	    x2 = x1 + (int) ((float) image->width * scale);
	    y2 = y1 + (int) ((float) image->height * scale);

	    glRotatef(frame->type_2.angle, 0.0, 0.0, 1.0);
	    
	    // unk6 seems to flip, as seen in reisen/tenshi backwalk?
	} else {
	    x1 = -(frame->header.x_offset);
	    y1 = -(frame->header.y_offset);
	    x2 = x1 + (image->width * 2);
	    y2 = y1 + (image->height * 2);
	}

	glBindTexture(GL_TEXTURE_2D, image->tex_id);
	glEnable(GL_TEXTURE_2D);

	glColor4f(1.0, 1.0, 1.0, 1.0);
	glBegin(GL_QUADS);
	glTexCoord2f(0.0, 0.0);
	glVertex2f(x1, y1);
	glTexCoord2f(image->x_mult, 0.0);
	glVertex2f(x2, y1);
	glTexCoord2f(image->x_mult, image->y_mult);
	glVertex2f(x2, y2);
	glTexCoord2f(0.0, image->y_mult);
	glVertex2f(x1, y2);
	glEnd();

	glDisable(GL_TEXTURE_2D);

	if (pglBlendEquation) {
	    pglBlendEquation(GL_FUNC_ADD);
	}
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glPopMatrix();
    }
}

void display_load()
{
    glFinish();

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    display_bg();

    glColor4f(0.0, 0.0, 0.0, 1.0);
    draw_font((window_width / 2) - (7 * 7) + 1, 21, "Loading");
    glColor4f(1.0, 1.0, 0.6, 1.0);
    draw_font((window_width / 2) - (7 * 7), 20, "Loading");

    if (ncharacters > 0) {
	character_t *ch = &characters[ncharacters - 1];
	if (ch->nsequences > 0) {
	    sequence_t *seq = &ch->sequences[0];
	    if (seq->header.nframes > 0) {
		frame_t *frame = &seq->frames[0];

		glPushMatrix();
		glTranslatef((window_width / 2), (window_height - 50),
			     0.0);
		glScalef(window_scale, window_scale, 1.0);

		display_sprite(ch, frame);

		glPopMatrix();
	    }
	}
    }


    int y = 60;
    for (int i = 0; i < ncharacters; ++i) {
	const char *name = characters[i].name;

	glColor4f(0.0, 0.0, 0.0, 1.0);
	draw_font((window_width / 2) - (10 * 7) + 1, y + 1, name);
	glColor4f(1.0, 1.0, 0.6, 1.0);
	draw_font((window_width / 2) - (10 * 7), y, name);

	y += 15;
    }

    SDL_GL_SwapBuffers();
}

void display_scene(character_t * character, frame_t * frame,
		   int ch, int seq, int nframe, int fcount)
{
    glFinish();

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    display_bg();

    if (draw_axises) {
	glColor4f(0.8, 0.8, 0.0, 1.0);
	glBegin(GL_LINES);
	glVertex2f(0.0, position_y);
	glVertex2f(window_width, position_y);
	glVertex2f(position_x, 0.0);
	glVertex2f(position_x, window_height);
	glEnd();
    }

    if (frame) {
	int off_x = position_x;
	int off_y = position_y;
	glPushMatrix();
	glTranslatef(off_x, off_y, 0.0);
	glScalef(window_scale, window_scale, 1.0);

	if (draw_sprite
	    && frame->header.render_frame >= 0
	    && frame->header.render_frame < character->nimages
	    && character->images[frame->header.render_frame].valid) {

	    display_sprite(character, frame);

	}
	if (pretty_display) {
	    if (draw_collisionbox && frame->n_rects > 0) {
		node *tmp = NULL;
		glColor4f(0.8, 0.8, 0.8, 0.25);
		do {
		    if (tmp == NULL) {
			tmp = create_box(frame, 2);
		    } else {
			node *tmp2 = tmp;
			tmp = tmp->next;
			delete tmp2;
		    }
		    glBegin(GL_QUADS);
		    glVertex2f(tmp->x1, tmp->y1);
		    glVertex2f(tmp->x2, tmp->y1);
		    glVertex2f(tmp->x2, tmp->y2);
		    glVertex2f(tmp->x1, tmp->y2);
		    glEnd();
		}
		while (tmp->next != NULL);
		delete tmp;
	    }
	    if (draw_hitbox && frame->n_selfboxes > 0) {
		node *tmp = NULL;
		glColor4f(0.3, 1.0, 0.3, 0.3);
		do {
		    if (tmp == NULL) {
			tmp = create_box(frame, 0);
		    } else {
			node *tmp2 = tmp;
			tmp = tmp->next;
			delete tmp2;
		    }
		    glBegin(GL_QUADS);
		    glVertex2f(tmp->x1, tmp->y1);
		    glVertex2f(tmp->x2, tmp->y1);
		    glVertex2f(tmp->x2, tmp->y2);
		    glVertex2f(tmp->x1, tmp->y2);
		    glEnd();
		}
		while (tmp->next != NULL);
		delete tmp;
	    }
	    if (draw_damagebox && frame->n_hitboxes > 0) {
		node *tmp = NULL;
		glColor4f(1.0, 0.3, 0.3, 0.3);
		do {
		    if (tmp == NULL) {
			tmp = create_box(frame, 1);
		    } else {
			node *tmp2 = tmp;
			tmp = tmp->next;
			delete tmp2;
		    }
		    glBegin(GL_QUADS);
		    glVertex2f(tmp->x1, tmp->y1);
		    glVertex2f(tmp->x2, tmp->y1);
		    glVertex2f(tmp->x2, tmp->y2);
		    glVertex2f(tmp->x1, tmp->y2);
		    glEnd();
		}
		while (tmp->next != NULL);
	    }
	}
	if (draw_boxes) {
	    if (draw_collisionbox) {
		glColor4f(0.9, 0.9, 0.9, 1.0);
		for (int i = 0; i < frame->n_rects; ++i) {
		    rect_t *rect = &frame->rects[i];
		    glBegin(GL_LINE_LOOP);
		    glVertex2f(rect->x1, rect->y1);
		    glVertex2f(rect->x2, rect->y1);
		    glVertex2f(rect->x2, rect->y2);
		    glVertex2f(rect->x1, rect->y2);
		    glEnd();
		}
	    }
	    if (draw_hitbox && frame->n_selfboxes > 0) {
		glColor4f(0.3, 1.0, 0.3, 1.0);
		for (int i = 0; i < frame->n_selfboxes; ++i) {
		    rect_t *rect = &frame->selfboxes[i];
		    glBegin(GL_LINE_LOOP);
		    glVertex2f(rect->x1, rect->y1);
		    glVertex2f(rect->x2, rect->y1);
		    glVertex2f(rect->x2, rect->y2);
		    glVertex2f(rect->x1, rect->y2);
		    glEnd();
		}
	    }
	    if (draw_damagebox && frame->n_hitboxes > 0) {
		glColor4f(1.0, 0.3, 0.3, 1.0);
		for (int i = 0; i < frame->n_hitboxes; ++i) {
		    rect_t *rect = &frame->hitboxes[i];
		    glBegin(GL_LINE_LOOP);
		    glVertex2f(rect->x1, rect->y1);
		    glVertex2f(rect->x2, rect->y1);
		    glVertex2f(rect->x2, rect->y2);
		    glVertex2f(rect->x1, rect->y2);
		    glEnd();
		}
	    }
	}

	glPopMatrix();

	if (draw_properties == 1) {
	    char strbuf[1024];
	    sprintf(strbuf,
		    "duration %d\n\n fflags %8.8x\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s\n\n\n",
		    frame->header.duration, frame->data.fflags,
		    (frame->data.fflags & 1) ? "Standing" : "-",
		    (frame->data.fflags & 2) ? "Crouching" : "-",
		    (frame->data.fflags & 4) ? "Airborne" : "-",
		    (frame->data.fflags & 16) ? "Guard cancel" : "-",
		    (frame->data.fflags & 32) ? "Cancellable" : "-",
		    (frame->data.fflags & 64) ? "CH when hit" : "-",
		    (frame->data.fflags & 128) ? "Superarmor" : "-",
		    // 256, ?, 0x44488f extended superarmor, unused
		    (frame->data.fflags & 8192) ? "Melee Inv" : "-",
		    (frame->data.fflags & 16384) ? "Graze" : "-");
	    // 32768, ?, 0x444fe7 related to af 8192
	    // 65536, ?, 0x445042 related to af 16384
	    // 131072, ?, 0x445099 related to af 32768
	    if (frame->n_hitboxes > 0) {
		sprintf(strbuf + strlen(strbuf),
			" aflags %8.8x\n block: %s %s %s %s\n%s\n%s\n%s\n%s\n\n",
			frame->data.aflags,
			(frame->data.aflags & 2) ? "H" : "-",
			(frame->data.aflags & 4) ? "L" : "-",
			(frame->data.aflags & 8) ? "A" : "-",
			(frame->data.aflags & 16) ? "U" : "-",
			// 32, ? 0x444863 related to ff 128/256 - ignore SA
			// only used by youmu throw
			// 64, ?
			// 128, ?
			// 256, ? 0x44499f, juggle related
			// 512, ?
			(frame->data.aflags & 1024) ? "Induces CH" : "-",
			// 8192, ? 0x444fd3 related to ff 32768
			// 16384, ? 0x44502e related to ff 65536
			// 32768, ? 0x445086 related to ff 131072
			(frame->data.
			 aflags & 262144) ? "Ungrazeable" : "-",
			(frame->data.
			 aflags & 524288) ? "Guard crush" : "-",
			// 1048576, "Always CH", unused
			(frame->data.aflags & 2097152) ? "Stagger" : "-");
		sprintf(strbuf + strlen(strbuf),
			"   Damage: %d\nProration: %d\n    Limit: %d\n   Untech: %d\n  Unknown: %d %d\nVelocityX: %d\nVelocityY: %d\n     Type: %d %d %d\n",
			frame->data.damage, frame->data.proration,
			frame->data.limit, frame->data.untech,
			frame->data.unk9, frame->data.unk10,
			frame->data.velocity_x, frame->data.velocity_y,
			frame->data.unk15, frame->data.unk16,
			frame->data.type);
	    }

	    glColor4f(0.0, 0.0, 0.0, 1.0);
	    draw_font(window_width - 159, 16, strbuf);
	    glColor4f(1.0, 1.0, 1.0, 1.0);
	    draw_font(window_width - 160, 15, strbuf);
	} else if (draw_properties == 2) {
	    char strbuf[1024];
	    sprintf(strbuf,
		    " Duration: %d\n      h 2: %d\n      h 3: %d\n      h 4: %d\n\n"
		    "   Damage: %d\nProration: %d\n"
		    "        3: %d\n        4: %d\n   Untech: %d\n        6: %d\n"
		    "    Limit: %d\n        8: %d\n        9: %d\n       10: %d\n"
		    "       11: %d\n       12: %d\n       13: %d\n       14: %d\n"
		    "       15: %d\nVelocityX: %d\nVelocityY: %d\n       18: %d\n"
		    "       19: %d\n     Type: %d\n       21: %d\n"
		    "   fflags: %8.8x\n   aflags: %8.8x\n",
		    frame->header.duration,
		    frame->header.unk2,
		    frame->header.unk3,
		    frame->header.unk4,
		    frame->data.damage,
		    frame->data.proration,
		    frame->data.unk3,
		    frame->data.unk4,
		    frame->data.untech,
		    frame->data.newunk5,
		    frame->data.limit,
		    frame->data.unk5,
		    frame->data.unk6,
		    frame->data.unk7,
		    frame->data.unk8,
		    frame->data.unk9,
		    frame->data.unk10,
		    frame->data.unk11,
		    frame->data.unk12,
		    frame->data.velocity_x,
		    frame->data.velocity_y,
		    frame->data.unk15,
		    frame->data.unk16,
		    frame->data.type,
		    frame->data.unk18,
		    frame->data.fflags, frame->data.aflags);

	    glColor4f(0.0, 0.0, 0.0, 1.0);
	    draw_font(window_width - 164, 16, strbuf);
	    glColor4f(1.0, 1.0, 1.0, 1.0);
	    draw_font(window_width - 165, 15, strbuf);
	}
    }

    if (character && draw_properties) {
	int current = 0;
	if (character->sequences[seq].header.nframes > 0) {
	    for (int i = 0; i < nframe; i++) {
		current +=
		    character->sequences[seq].frames[i].header.duration;
	    }
	    fcount += current;
	    for (int i = nframe;
		 i < character->sequences[seq].header.nframes; i++) {
		current +=
		    character->sequences[seq].frames[i].header.duration;;
	    }
	}

	char stringbuf[256];
	sprintf(stringbuf,
		"%3d %-10s seq: %4d-%4d fr: %5d-%5d   sp: %3d-%3d   ID: %3d",
		alt_colors,
		character->name, seq + 1, character->nsequences,
		fcount + 1, current, nframe + 1,
		character->sequences[seq].header.nframes,
		character->sequences[seq].identifier);

	glColor4f(0.0, 0.0, 0.0, 1.0);
	draw_font(window_width - 590, window_height - 20, stringbuf);
	glColor4f(1.0, 1.0, 0.6, 1.0);
	draw_font(window_width - 591, window_height - 21, stringbuf);
    }
}

void setup_opengl()
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, window_width, window_height, 0, -10, 10);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    init_font();
}

int main(int argc, char **argv)
{
    datfile_t *th105a_dat, *th105c_dat, *th123a_dat, *th123c_dat;
    
    char base_filename[1024];
    char filename[1536];
    
    strcpy(base_filename, "");
    
#ifdef WIN32
    GetPrivateProfileString("th105path", "path", "", base_filename, 1023, ".\\configex123.ini");
    if (strlen(base_filename) > 0) {
        strcat(base_filename, "\\");
    }
#endif
    
    sprintf(filename, "%sth105a.dat", base_filename);
    th105a_dat = dat_open(filename);
    th123a_dat = dat_open("th123a.dat");
    
    if (!th105a_dat && !th123a_dat) {
	printf("couldn't open th105a or th123a\n");
	return -1;
    }

    sprintf(filename, "%sth105c.dat", base_filename);
    th105c_dat = dat_open(filename);
    th123c_dat = dat_open("th123c.dat");

    SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER);

    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
    SDL_Surface *surface =
	SDL_SetVideoMode(640, 480, 0, SDL_OPENGL | SDL_RESIZABLE);
    if (!surface) {
	return -1;
    }

    SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY,
			SDL_DEFAULT_REPEAT_INTERVAL);

    int ch = 0;
    int seq = 0;
    int frame = 0;

    setup_opengl();

    pglBlendEquation =
	(PFNGLBLENDEQUATIONEXTPROC)
	SDL_GL_GetProcAddress("glBlendEquation");

    SDL_WM_SetCaption("framedisplay 1.7 for SWR & Soku", 0);

    display_load();

    read_pats(th105a_dat, th123a_dat, th105c_dat, th123c_dat);

    display_scene(0, 0, ch, seq, frame, 0);

    SDL_GL_SwapBuffers();

    int done = 0, render = 1, animate = 0, frame_count = 0;
    int ms_down = 0, ms_x = 0, ms_y = 0, start_x = 0, start_y = 0;
    while (!done) {
	if (animate && characters[ch].sequences[seq].header.nframes > 0) {
	    frame_count++;
	    if (frame_count >=
		characters[ch].sequences[seq].frames[frame].header.
		duration) {
		frame++;
		if (frame >= characters[ch].sequences[seq].header.nframes) {
		    frame = 0;
		}

		frame_count = 0;

		render = 1;
	    }
	}
	if (render) {
	    if (characters[ch].sequences[seq].header.nframes > 0) {
		display_scene(&characters[ch],
			      &characters[ch].sequences[seq].frames[frame],
			      ch, seq, frame, frame_count);
	    } else {
		display_scene(&characters[ch], 0, ch, seq, frame,
			      frame_count);
	    }
	    SDL_GL_SwapBuffers();

	    render = 0;
	}

	SDL_Delay(16);

	SDL_Event sdl_event;
	SDL_PumpEvents();

	while (SDL_PollEvent(&sdl_event)) {
	    switch (sdl_event.type) {
	    case SDL_KEYDOWN:
		switch (sdl_event.key.keysym.sym) {
		case SDLK_ESCAPE:
		    done = 1;
		    break;
		case SDLK_F1:
		    draw_gradient = 1 - draw_gradient;
		    render = 1;
		    break;
		case SDLK_F2:
		    draw_properties = (draw_properties + 1) % 3;
		    render = 1;
		    break;
		case SDLK_1:
		    draw_sprite = 1 - draw_sprite;
		    render = 1;
		    break;
		case SDLK_2:
		    draw_collisionbox = 1 - draw_collisionbox;
		    render = 1;
		    break;
		case SDLK_3:
		    draw_hitbox = 1 - draw_hitbox;
		    render = 1;
		    break;
		case SDLK_4:
		    draw_damagebox = 1 - draw_damagebox;
		    render = 1;
		    break;
		case SDLK_5:
		    draw_additive = 1 - draw_additive;
		    render = 1;
		    break;
		case SDLK_6:
		    pretty_display = 1 - pretty_display;
		    render = 1;
		    break;
		case SDLK_7:
		    draw_boxes = 1 - draw_boxes;
		    render = 1;
		    break;
		case SDLK_8:
		    realframes = 1 - realframes;
		    render = 1;
		    break;
		case SDLK_9:
		    draw_axises = 1 - draw_axises;
		    render = 1;
		    break;
		case SDLK_SPACE:
		    animate = 1 - animate;
		    frame_count = 0;
		    break;
		case SDLK_UP:
		case SDLK_KP8:
		    frame_count = 0;
		    seq--;
		    frame = 0;
		    if (seq < 0) {
			seq = characters[ch].nsequences - 1;
		    }
		    render = 1;
		    break;
		case SDLK_DOWN:
		case SDLK_KP2:
		    frame_count = 0;
		    seq++;
		    frame = 0;
		    if (seq >= characters[ch].nsequences) {
			seq = 0;
		    }
		    render = 1;
		    break;
		case SDLK_LEFT:
		case SDLK_KP4:
		    if (realframes == 0) {
			frame_count = 0;
			frame--;
			if (frame < 0) {
			    frame =
				characters[ch].sequences[seq].header.
				nframes - 1;
			}
		    } else {
			frame_count--;
			if (frame_count < 0) {
			    frame--;
			    if (frame < 0) {
				frame =
				    characters[ch].sequences[seq].header.
				    nframes - 1;
			    }
			    frame_count =
				characters[ch].sequences[seq].
				frames[frame].header.duration - 1;
			}
		    }
		    render = 1;
		    break;
		case SDLK_RIGHT:
		case SDLK_KP6:
		    if (realframes == 0) {
			frame_count = 0;
			frame++;
			if (frame >=
			    characters[ch].sequences[seq].header.nframes) {
			    frame = 0;
			}
		    } else {
			frame_count++;
			if (frame_count >
			    characters[ch].sequences[seq].frames[frame].
			    header.duration - 1) {
			    frame++;
			    if (frame >=
				characters[ch].sequences[seq].header.
				nframes) {
				frame = 0;
			    }
			    frame_count = 0;
			}
		    }
		    render = 1;
		    break;
		case SDLK_PAGEUP:
		case SDLK_KP9:
		    frame_count = 0;
		    ch--;
		    seq = 0;
		    frame = 0;
		    if (ch < 0) {
			ch = ncharacters - 1;
		    }
		    render = 1;
		    break;
		case SDLK_PAGEDOWN:
		case SDLK_KP3:
		    frame_count = 0;
		    ch++;
		    seq = 0;
		    frame = 0;
		    if (ch >= ncharacters) {
			ch = 0;
		    }
		    render = 1;
		    break;
		case SDLK_MINUS:
		case SDLK_KP_MINUS:
		    window_scale = window_scale / 1.1;
		    render = 1;
		    break;
		case SDLK_PLUS:
		case SDLK_EQUALS:
		case SDLK_KP_PLUS:
		    window_scale = window_scale * 1.1;
		    render = 1;
		    break;
		case SDLK_BACKSPACE:
		    window_scale = 1.0;
		    render = 1;
		    break;
		case SDLK_TAB:
		    for (int i = 0; i < 10; ++i) {
		        alt_colors = (alt_colors+1)%10;
		        if (characters[ch].palettes[alt_colors].active) {
		            break;
		        }
                    }
		    flush_images();
		    render = 1;
		    break;
		default:
		    break;
		}

		break;
	    case SDL_MOUSEBUTTONDOWN:
		ms_down = 1;
		ms_x = sdl_event.button.x;
		ms_y = sdl_event.button.y;
		start_x = position_x;
		start_y = position_y;
		break;
	    case SDL_MOUSEBUTTONUP:
		ms_down = 0;
		break;
	    case SDL_MOUSEMOTION:
		if (ms_down == 1) {
		    position_x = start_x + sdl_event.motion.x - ms_x;
		    position_y = start_y + sdl_event.motion.y - ms_y;
		    render = 1;
		}
		break;
	    case SDL_VIDEOEXPOSE:
		render = 1;
		break;
	    case SDL_VIDEORESIZE:
		SDL_SetVideoMode(sdl_event.resize.w, sdl_event.resize.h, 0,
				 SDL_OPENGL | SDL_RESIZABLE);
		glViewport(0, 0, sdl_event.resize.w, sdl_event.resize.h);

		window_width = sdl_event.resize.w;
		window_height = sdl_event.resize.h;

		setup_opengl();

		current_tex_cycle++;
		break;
	    case SDL_QUIT:
		done = 1;
		break;
	    }
	}
    }

    SDL_Quit();

    return 0;
}
